{# djlint:off H030,H031 #}
{% block doc -%}
<!DOCTYPE html>
<html{% block html_attribs %}{% endblock html_attribs %}>
  {%- block html %}
  <head>
    {%- block head %}
    <title>{% block title %}{{ title|default }}{% endblock title %}</title>

      {%- block metas %}
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
      {%- endblock metas %}

      {%- block head_scripts %}
      {%- endblock head_scripts %}

      {%- block styles %}
        <style>
          .fs-center { text-align: center }
          .fs-important { font-size: larger; font-weight: bold }
          .fs-gap { margin-top: 20px; }
          .fs-div { margin: 4px; }
          .fs-error-msg { color: darkred; }
        </style>
      {%- endblock styles %}
    {%- endblock head %}
  </head>
  <body{% block body_attribs %}{% endblock body_attribs %}>
    {% block body -%}
      {% block navbar %}
      {%- endblock navbar %}
      {% block content -%}
      {%- endblock content %}

      {% block scripts %}
      {%- endblock scripts %}
    {%- endblock body %}
  </body>
  {%- endblock html %}
</html>
{% endblock doc -%}
