{% set title = title|default(_fsdomain("Enter Recovery Code")) %}
{% extends "security/base.html" %}
{% from "security/_macros.html" import render_field_with_errors, render_field %}

{% block content %}
  {% include "security/_messages.html" %}
  <h1>{{ _fsdomain("Enter Recovery Code") }}</h1>
  <form action="{{ url_for_security('mf_recovery') }}" method="post" name="mf_recovery_form">
    {{ mf_recovery_form.hidden_tag() }}
    {{ render_field_with_errors(mf_recovery_form.code) }}
    {{ render_field(mf_recovery_form.submit) }}
  </form>
  {% include "security/_menu.html" %}
{% endblock content %}
