{# This template receives the following context:
  login_link - the link that should be fetched (GET) to reset
  login_token - this token is part of reset link - but can be used to
    construct arbitrary URLs for redirecting.
  user - the entire user model object
  username - username
  security - the Flask-Security configuration
#}
<p>{{ _fsdomain("Welcome %(username)s!", username=username) }}</p>
<p>{{ _fsdomain("You can sign in to your account using the following code: %(token)s", token=token) }}</p>
{% if login_link %}
  <p>{{ _fsdomain('Or use this link: <a href="%(login_link)s">Sign in</a>', login_link=login_link)|safe }}</p>
{% endif %}
